# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

from awkward._v2.types.type import Type  # noqa: F401
from awkward._v2.types.unknowntype import UnknownType  # noqa: F401
from awkward._v2.types.numpytype import NumpyType  # noqa: F401
from awkward._v2.types.regulartype import RegularType  # noqa: F401
from awkward._v2.types.listtype import ListType  # noqa: F401
from awkward._v2.types.recordtype import RecordType  # noqa: F401
from awkward._v2.types.optiontype import OptionType  # noqa: F401
from awkward._v2.types.uniontype import UnionType  # noqa: F401
from awkward._v2.types.arraytype import ArrayType  # noqa: F401
